/*
  Simple DirectMedia Layer
  Copyright (C) 1997-2022 Sam Lantinga <slouken@libsdl.org>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/

/**
 *  \file SDL_video_capture.h
 *
 *  Video Capture for the SDL library.
 */

#ifndef SDL_video_capture_h_
#define SDL_video_capture_h_

#include "SDL3/SDL_video.h"

#include <SDL3/SDL_begin_code.h>
/* Set up for C function definitions, even when using C++ */
#ifdef __cplusplus
extern "C" {
#endif


/**
 *  SDL VideoCapture Device IDs.
 */
typedef Sint32 SDL_VideoCaptureDeviceID;
#define SDL_VIDEO_CAPTURE_INVALID                  -1

#define SDL_VIDEO_CAPTURE_ALLOW_ANY_CHANGE          1

/**
 *  SDL_VideoCaptureSpec structure
 *
 *  Only those field can be 'desired' when configuring the device:
 *  - format
 *  - width
 *  - height
 *
 *  \sa SDL_GetVideoCaptureFormat
 *  \sa SDL_GetVideoCaptureFrameSize
 *
 */
typedef struct SDL_VideoCaptureSpec
{
    Uint32 format;          /**< Video data format */
    Uint32 width;           /**< Width */
    Uint32 height;          /**< Height */
} SDL_VideoCaptureSpec;

/**
 *  SDL Video Capture Status
 */
typedef enum
{
    SDL_VIDEO_CAPTURE_INIT = 0, /**< Init, spec hasn't been set */
    SDL_VIDEO_CAPTURE_STOPPED,  /**< Stopped */
    SDL_VIDEO_CAPTURE_PLAYING   /**< Playing */
} SDL_VideoCaptureStatus;

/**
 *  SDL Video Capture Status
 */
typedef struct SDL_VideoCaptureFrame
{
    int num_planes;             /**< Number of planes */
    Uint8 *data[3];              /**< Pointer to data of i-th plane */
    int row_stride[3];          /**< Row stride of i-th plane */
    int fd;                     /**< DMABUF file descriptor if available (-1 if not) */
    void *clientbuffer;         /**< Android clientbuffer if available (NULL if not) */
    void *internal;             /**< Private field */
} SDL_VideoCaptureFrame;

/**
 * Open a Video Capture device
 *
 * \param device_name device name to open. Can be NULL.
 *
 * \returns device id, or SDL_VIDEO_CAPTURE_INVALID on error
 *
 * \sa SDL_GetVideoCaptureDeviceName
 * \sa SDL_GetNumVideoCaptureDevices
 */
extern DECLSPEC SDL_VideoCaptureDeviceID SDLCALL SDL_OpenVideoCapture(const char *device_name);

/**
 * Set specification
 *
 * \param id video capture opened device id
 * \param desired desired video capture spec
 * \param obtained obtained video capture spec
 * \param allowed_changes allow changes or not
 *
 * \returns 0 on success, negative value on error
 */
extern DECLSPEC int SDLCALL SDL_SetVideoCaptureSpec(SDL_VideoCaptureDeviceID id,
                                                  const SDL_VideoCaptureSpec *desired,
                                                  SDL_VideoCaptureSpec *obtained,
                                                  int allowed_changes);

/**
 * Open a Video Capture device and set specification
 *
 * \param device_name device name to open. Can be NULL.
 * \param desired desired video capture spec
 * \param obtained obtained video capture spec
 * \param allowed_changes allow changes or not
 *
 * \returns device id, or SDL_VIDEO_CAPTURE_INVALID on error
 */
extern DECLSPEC SDL_VideoCaptureDeviceID SDLCALL SDL_OpenVideoCaptureWithSpec(
                                                  const char *device_name,
                                                  const SDL_VideoCaptureSpec *desired,
                                                  SDL_VideoCaptureSpec *obtained,
                                                  int allowed_changes);
/**
 * Get device name
 *
 * \param index device index between 0 and num -1
 *
 * \returns device name, shouldn't be freed
 *
 * \sa SDL_GetNumVideoCaptureDevices
 */
extern DECLSPEC const char * SDLCALL SDL_GetVideoCaptureDeviceName(int index);

/**
 * Use this function to get the number of video capture devices.
 *
 * \returns the number of video capture devices
 */
extern DECLSPEC int SDLCALL SDL_GetNumVideoCaptureDevices(void);

/**
 * Get the obtained video capture spec
 *
 * \param id video capture opened device id
 * \param spec The SDL_VideoCaptureSpec to be initialized by this function.
 *
 * \returns 0 on success, negative value on error
 */
extern DECLSPEC int SDLCALL SDL_GetVideoCaptureSpec(SDL_VideoCaptureDeviceID id, SDL_VideoCaptureSpec *spec);


#define SDL_VIDEO_CAPTURE_TYPE_COMPRESSED   1
#define SDL_VIDEO_CAPTURE_TYPE_EMULATED     2
/**
 * Get frame format of video capture device.
 * The value can be used to fill SDL_VideoCaptureSpec structure.
 *
 * \param id video capture opened device id
 * \param index format between 0 and num -1
 * \param format pointer output format (SDL_PixelFormatEnum)
 * \param type pointer output type
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_GetNumVideoCaptureFormats
 */
extern DECLSPEC int SDLCALL SDL_GetVideoCaptureFormat(SDL_VideoCaptureDeviceID id, int index, Uint32 *format, Uint32 *type);

/**
 * Number of available formats for the device
 *
 * \param id video capture opened device id
 *
 * \returns number of formats, negative value on error
 *
 * \sa SDL_GetVideoCaptureFormat
 */
extern DECLSPEC int SDLCALL SDL_GetNumVideoCaptureFormats(SDL_VideoCaptureDeviceID id);

/**
 * Get frame sizes of the device and the specified input format.
 * The value can be used to fill SDL_VideoCaptureSpec structure.
 *
 * \param id video capture opened device id
 * \param format a format that can be used by the device (SDL_PixelFormatEnum)
 * \param index framesize between 0 and num -1
 * \param width output width
 * \param height output height
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_GetNumVideoCaptureFrameSizes
 */
extern DECLSPEC int SDLCALL SDL_GetVideoCaptureFrameSize(SDL_VideoCaptureDeviceID id, Uint32 format, int index, Uint32 *width, Uint32 *height);

/**
 * Number of different framesizes available for the device and pixel format.
 *
 * \param id video capture opened device id
 * \param format frame pixel format (SDL_PixelFormatEnum)
 *
 * \returns number of framesize, negative value on error
 *
 * \sa SDL_GetVideoCaptureFrameSize
 */
extern DECLSPEC int SDLCALL SDL_GetNumVideoCaptureFrameSizes(SDL_VideoCaptureDeviceID id, Uint32 format);


/**
 * Get video capture status
 *
 * \param id video capture opened device id
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_VideoCaptureStatus
 */
extern DECLSPEC SDL_VideoCaptureStatus SDLCALL SDL_GetVideoCaptureStatus(SDL_VideoCaptureDeviceID id);

/**
 * Start video capture
 *
 * \param id video capture opened device id
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_StopVideoCapture
 */
extern DECLSPEC int SDLCALL SDL_StartVideoCapture(SDL_VideoCaptureDeviceID id);

/**
 * Acquire a frame.
 * The frame is a memory pointer to the image data, whose size and format
 * are given by the the obtained spec.
 *
 * Non blocking API. If there is a frame available, frame->num_planes is non 0.
 * If frame->num_planes is 0 and returned code is 0, there is no frame at that time.
 *
 * After used, the frame should be released with SDL_VideoCaptureReleaseFrame
 *
 * \param id video capture opened device id
 * \param frame pointer to get the frame
 * \param ticks pointer to get the ticks when this was read from driver
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_VideoCaptureReleaseFrame
 */
extern DECLSPEC int SDLCALL SDL_VideoCaptureAcquireFrame(SDL_VideoCaptureDeviceID id, SDL_VideoCaptureFrame *frame, Uint64 *ticks);

/**
 * Release a frame. Let the back-end re-use the internal buffer for video capture.
 *
 * \param id video capture opened device id
 * \param frame frame pointer.
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_VideoCaptureAcquireFrame
 */
extern DECLSPEC int SDLCALL SDL_VideoCaptureReleaseFrame(SDL_VideoCaptureDeviceID id, SDL_VideoCaptureFrame *frame);

/**
 * Stop Video Capture
 *
 * \param id video capture opened device id
 *
 * \returns 0 on success, negative value on error
 *
 * \sa SDL_StartVideoCapture
 */
extern DECLSPEC int SDLCALL SDL_StopVideoCapture(SDL_VideoCaptureDeviceID id);

/**
 * Use this function to shut down video_capture processing and close the video_capture device.
 *
 * The device ID is invalid as soon as the device is closed, and is eligible
 * for reuse in a new SDL_OpenVideoCaptureWithSpec() call immediately.
 *
 * \param id video capture opened device id
 *
 * \sa SDL_OpenVideoCaptureWithSpec
 * \sa SDL_OpenVideoCapture
 */
extern DECLSPEC void SDLCALL SDL_CloseVideoCapture(SDL_VideoCaptureDeviceID id);

/* Ends C function definitions when using C++ */
#ifdef __cplusplus
}
#endif
#include <SDL3/SDL_close_code.h>

#endif /* SDL_video_capture_h_ */
