/*
  Simple DirectMedia Layer
  Copyright (C) 1997-2022 Sam Lantinga <slouken@libsdl.org>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/
#include "../SDL_internal.h"

#include "SDL3/SDL.h"
#include "SDL3/SDL_video_capture.h"
#include "SDL_sysvideocapture.h"
#include "SDL_video_capture_c.h"
#include "../thread/SDL_systhread.h"

#define DEBUG_VIDEO_CAPTURE_CAPTURE 1


/* list node entries to share frames between SDL and user app */
typedef struct entry_t
{
    SDL_VideoCaptureFrame frame;
    Uint64 ticks;
} entry_t;

static SDL_VideoCaptureDevice *open_devices[16];

static SDL_VideoCaptureDevice *
get_video_capture_device(SDL_VideoCaptureDeviceID id)
{
    Uint32 i = id;
    if (i >= SDL_arraysize(open_devices) -1 || open_devices[i] == NULL) {
        SDL_SetError("Invalid video capture device ID");
        return NULL;
    }
    return open_devices[i];
}

static void
close_device(SDL_VideoCaptureDevice *device)
{
    if (!device) {
        return;
    }

    SDL_AtomicSet(&device->shutdown, 1);
    SDL_AtomicSet(&device->enabled, 1);

    if (device->thread != NULL) {
        SDL_WaitThread(device->thread, NULL);
    }
    if (device->device_lock != NULL) {
        SDL_DestroyMutex(device->device_lock);
    }
    if (device->acquiring_lock != NULL) {
        SDL_DestroyMutex(device->acquiring_lock);
    }

    open_devices[device->id] = NULL;

    CloseDevice(device);

    {
        entry_t *entry = NULL;
        while (device->buffer_queue != NULL) {
            SDL_ListPop(&device->buffer_queue, (void**)&entry);
            if (entry) {
                SDL_free(entry);
            }
        }
    }

    SDL_free(device->dev_name);
    SDL_free(device);
}

#if __IPHONEOS__ || __MACOS__
/* unused */
#else
static void
size_buffer(Uint32 width, Uint32 height, Uint32 format, Uint32 *size, Uint32 *pitch)
{
   const int sz_plane         = width * height;
   const int sz_plane_chroma  = ((width + 1) / 2) * ((height + 1) / 2);
   const int sz_plane_packed  = ((width + 1) / 2) * height;

   switch (format)
   {
      case SDL_PIXELFORMAT_YV12: /**< Planar mode: Y + V + U  (3 planes) */
      case SDL_PIXELFORMAT_IYUV: /**< Planar mode: Y + U + V  (3 planes) */
         *size = sz_plane + sz_plane_chroma + sz_plane_chroma;
         *pitch = width;
         break;

      case SDL_PIXELFORMAT_YUY2: /**< Packed mode: Y0+U0+Y1+V0 (1 plane) */
      case SDL_PIXELFORMAT_UYVY: /**< Packed mode: U0+Y0+V0+Y1 (1 plane) */
      case SDL_PIXELFORMAT_YVYU: /**< Packed mode: Y0+V0+Y1+U0 (1 plane) */
         *size = 4 * sz_plane_packed;
         *pitch = ((width + 1)/2) * 4;
         break;

      case SDL_PIXELFORMAT_NV12: /**< Planar mode: Y + U/V interleaved  (2 planes) */
      case SDL_PIXELFORMAT_NV21: /**< Planar mode: Y + V/U interleaved  (2 planes) */
         *size = sz_plane + sz_plane_chroma + sz_plane_chroma;
         *pitch = width;
         break;

      default:
         {
            int bpp;
            Uint32 Rmask, Gmask, Bmask, Amask;
            SDL_GetMasksForPixelFormatEnum(format, &bpp, &Rmask, &Gmask, &Bmask, &Amask);
            *size = width * height * (bpp / 8);
            *pitch = width * (bpp / 8);
         }
   }
}
#endif

void
SDL_CloseVideoCapture(SDL_VideoCaptureDeviceID id)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        SDL_InvalidParamError("id");
        return;
    }
    close_device(device);
}

int
SDL_StartVideoCapture(SDL_VideoCaptureDeviceID id)
{
    SDL_VideoCaptureStatus status;
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        return SDL_InvalidParamError("id");
    }

    if (device->is_spec_set == SDL_FALSE) {
        return SDL_SetError("no spec set");
    }

    status = SDL_GetVideoCaptureStatus(id);
    if (status != SDL_VIDEO_CAPTURE_INIT) {
        return SDL_SetError("invalid state");
    }

    if (StartCapture(device) < 0) {
        return -1;
    }

    SDL_AtomicSet(&device->enabled, 1);

    return 0;
}

int
SDL_GetVideoCaptureSpec(SDL_VideoCaptureDeviceID id, SDL_VideoCaptureSpec *spec)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        return SDL_InvalidParamError("id");
    }

    if (!spec) {
        return SDL_InvalidParamError("spec");
    }

    SDL_zerop(spec);

    return GetDeviceSpec(device, spec);
}

int
SDL_StopVideoCapture(SDL_VideoCaptureDeviceID id)
{
    SDL_VideoCaptureStatus status;
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    int ret;
    if (!device) {
        return SDL_InvalidParamError("id");
    }

    status = SDL_GetVideoCaptureStatus(id);

    if (status != SDL_VIDEO_CAPTURE_PLAYING) {
        return SDL_SetError("invalid state");
    }

    SDL_AtomicSet(&device->enabled, 0);
    SDL_AtomicSet(&device->shutdown, 1);

    SDL_LockMutex(device->acquiring_lock);
    ret = StopCapture(device);
    SDL_UnlockMutex(device->acquiring_lock);

    if (ret < 0) {
        return -1;
    }

    return 0;
}

/* Check spec has valid format and frame size */
static int
prepare_video_capturespec(SDL_VideoCaptureDeviceID id, const SDL_VideoCaptureSpec *desired, SDL_VideoCaptureSpec *obtained, int allowed_changes)
{
    /* Check format */
    {
        int i, num = SDL_GetNumVideoCaptureFormats(id);
        int is_format_valid = 0;

        for (i = 0; i < num; i++) {
            Uint32 format, type;
            if (SDL_GetVideoCaptureFormat(id, i, &format, &type) == 0) {
                if (format == desired->format && format != SDL_PIXELFORMAT_UNKNOWN) {
                    is_format_valid = 1;
                    obtained->format = format;
                    break;
                }
            }
        }

        if (!is_format_valid) {
            if (allowed_changes) {
                /* Put first non compressed format, eventually emulated, better not emulated */
                for (i = 0; i < num; i++) {
                    Uint32 format, type;
                    if (SDL_GetVideoCaptureFormat(id, i, &format, &type) == 0) {
                        if (type & SDL_VIDEO_CAPTURE_TYPE_COMPRESSED) {
                            continue;
                        }
                        if (format != SDL_PIXELFORMAT_UNKNOWN) {
                            obtained->format = format;
                            is_format_valid = 1;
                        }
                        if ((type & SDL_VIDEO_CAPTURE_TYPE_EMULATED) == 0) {
                            break;
                        }
                    }
                }

            } else {
                SDL_SetError("Not allowed to change the format");
                return -1;
            }
        }

        if (!is_format_valid) {
            SDL_SetError("Invalid format");
            return -1;
        }
    }

    /* Check frame size */
    {
        int i, num = SDL_GetNumVideoCaptureFrameSizes(id, obtained->format);
        int is_framesize_valid = 0;

        for (i = 0; i < num; i++) {
            Uint32 w, h;
            if (SDL_GetVideoCaptureFrameSize(id, obtained->format, i, &w, &h) == 0) {
                if (desired->width == w && desired->height == h) {
                    is_framesize_valid = 1;
                    obtained->width = w;
                    obtained->height = h;
                    break;
                }
            }
        }

        if (!is_framesize_valid) {
            if (allowed_changes) {
                Uint32 w, h;
                if (SDL_GetVideoCaptureFrameSize(id, obtained->format, 0, &w, &h) == 0) {
                    is_framesize_valid = 1;
                    obtained->width = w;
                    obtained->height = h;
                }
            } else {
                SDL_SetError("Not allowed to change the frame size");
                return -1;
            }
        }

        if (!is_framesize_valid) {
            SDL_SetError("Invalid frame size");
            return -1;
        }

    }

    return 0;
}

const char *
SDL_GetVideoCaptureDeviceName(int index)
{
    static char buf[256];
    buf[0] = 0;
    buf[255] = 0;
    if (GetDeviceName(index, buf, sizeof (buf)) < 0) {
        buf[0] = 0;
    }
    return buf;
}

int
SDL_GetNumVideoCaptureDevices(void)
{
    return GetNumDevices();
}

/* Video capture thread function */
static int SDLCALL
SDL_CaptureVideoThread(void *devicep)
{
    const Uint32 delay = 20;
    SDL_VideoCaptureDevice *device = (SDL_VideoCaptureDevice *) devicep;

#if DEBUG_VIDEO_CAPTURE_CAPTURE
    SDL_Log("Start thread 'SDL_CaptureVideo'");
#endif


#if SDL_VIDEO_DRIVER_ANDROID
#if 0
    // TODO
    {
        /* Set thread priority to THREAD_PRIORITY_VIDEO */
        extern void Android_JNI_VideoCaptureSetThreadPriority(int, int);
        Android_JNI_VideoCaptureSetThreadPriority(device->iscapture, device->id);
    }
#endif
#else
    /* The video_capture mixing is always a high priority thread */
    SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH);
#endif

    /* Perform any thread setup */
    device->threadid = SDL_ThreadID();

    /* Init state */
    while (!SDL_AtomicGet(&device->enabled)) {
        SDL_Delay(delay);
    }

    /* Loop, filling the video_capture buffers */
    while (!SDL_AtomicGet(&device->shutdown)) {
        Uint64 t = 0;
        SDL_VideoCaptureFrame f;
        int ret;
        entry_t *entry;

        SDL_zero(f);
        f.fd = -1;
        f.clientbuffer = NULL;

        SDL_LockMutex(device->acquiring_lock);
        ret = AcquireFrame(device, &f, &t);
        SDL_UnlockMutex(device->acquiring_lock);

        if (ret == 0) {
            if (f.num_planes == 0) {
                continue;
            }
        }

        if (ret < 0) {
            /* Flag it as an error */
#if DEBUG_VIDEO_CAPTURE_CAPTURE
            SDL_Log("error AcquireFrame: %d %s",ret, SDL_GetError());
#endif
            f.num_planes = 0;
            t = 0;
        }


        entry = SDL_malloc(sizeof (entry_t));
        if (entry == NULL) {
            goto error_mem;
        }

        entry->frame = f;
        entry->ticks = t;

        SDL_LockMutex(device->device_lock);
        ret = SDL_ListAdd(&device->buffer_queue, entry);
        SDL_UnlockMutex(device->device_lock);

        if (ret < 0) {
            SDL_free(entry);
            goto error_mem;
        }
    }

#if DEBUG_VIDEO_CAPTURE_CAPTURE
    SDL_Log("End thread 'SDL_CaptureVideo'");
#endif
    return 0;

error_mem:
#if DEBUG_VIDEO_CAPTURE_CAPTURE
    SDL_Log("End thread 'SDL_CaptureVideo' with error: %s", SDL_GetError());
#endif
    SDL_AtomicSet(&device->shutdown, 1);
    SDL_OutOfMemory();
    return 0;
}

SDL_VideoCaptureDeviceID
SDL_OpenVideoCapture(const char *device_name)
{
    Uint32 i;
    SDL_VideoCaptureDeviceID id = SDL_VIDEO_CAPTURE_INVALID;
    SDL_VideoCaptureDevice *device = NULL;

    if (!SDL_WasInit(SDL_INIT_VIDEO)) {
        SDL_SetError("Video subsystem is not initialized");
        goto error;
    }

    /* !!! FIXME: there is a race condition here if two devices open from two threads at once. */
    /* Find an available device ID... */
    for (i = 0; i < SDL_arraysize(open_devices); i++) {
        if (open_devices[i] == NULL) {
            id = i;
            break;
        }
    }

    if (id == SDL_VIDEO_CAPTURE_INVALID) {
        SDL_SetError("Too many open video capture devices");
        goto error;
    }

    /* If app doesn't care about a specific device, let the user override. */
    if (device_name == NULL) {
        device_name = SDL_getenv("SDL_VIDEO_CAPTURE_DEVICE_NAME");
    }

    if (device_name == NULL) {
        device_name = SDL_GetVideoCaptureDeviceName(0);
    }

    if (device_name == NULL) {
        SDL_InvalidParamError("device");
        goto error;
    }

    device = (SDL_VideoCaptureDevice *) SDL_calloc(1, sizeof (SDL_VideoCaptureDevice));
    if (device == NULL) {
        SDL_OutOfMemory();
        goto error;
    }
    device->id = id;
    device->dev_name = SDL_strdup(device_name);


    SDL_AtomicSet(&device->shutdown, 0);
    SDL_AtomicSet(&device->enabled, 0);

    device->device_lock = SDL_CreateMutex();
    if (device->device_lock == NULL) {
        SDL_SetError("Couldn't create acquiring_lock");
        goto error;
    }

    device->acquiring_lock = SDL_CreateMutex();
    if (device->acquiring_lock == NULL) {
        SDL_SetError("Couldn't create acquiring_lock");
        goto error;
    }

    if (OpenDevice(device) < 0) {
        goto error;
    }

    /* empty */
    device->buffer_queue = NULL;
    open_devices[id] = device;  /* add it to our list of open devices. */


    /* Start the video_capture thread if necessary */
//    if (!current_video_capture.impl.ProvidesOwnCallbackThread) {
    if (1) {
        /* Start the video_capture thread */
        /* !!! FIXME: we don't force the video_capture thread stack size here if it calls into user code, but maybe we should? */
        /* buffer queueing callback only needs a few bytes, so make the stack tiny. */
        int is_internal_thread = 1;
        const size_t stacksize = is_internal_thread ? 64 * 1024 : 0;
        char threadname[64];

        SDL_snprintf(threadname, sizeof (threadname), "SDLVideoC%d", (int) device->id);
        device->thread = SDL_CreateThreadInternal(SDL_CaptureVideoThread, threadname, stacksize, device);

        if (device->thread == NULL) {
            SDL_SetError("Couldn't create video_capture thread");
            goto error;
        }
    }

    return device->id;

error:
    close_device(device);
    return SDL_VIDEO_CAPTURE_INVALID;
}

int
SDL_SetVideoCaptureSpec(SDL_VideoCaptureDeviceID id,
        const SDL_VideoCaptureSpec *desired,
        SDL_VideoCaptureSpec *obtained,
        int allowed_changes)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    SDL_VideoCaptureSpec _obtained;
    SDL_VideoCaptureSpec _desired;

    if (!device) {
        return SDL_InvalidParamError("id");
    }

    if (device->is_spec_set == SDL_TRUE) {
        return SDL_SetError("already configured");
    }

    if (!desired) {
        SDL_zero(_desired);
        desired = &_desired;
        allowed_changes = SDL_VIDEO_CAPTURE_ALLOW_ANY_CHANGE;
    } else {
        /* in case desired == obtained */
        _desired = *desired;
        desired = &_desired;
    }

    if (!obtained) {
        obtained = &_obtained;
    }

    SDL_zerop(obtained);

    if (prepare_video_capturespec(id, desired, obtained, allowed_changes) < 0) {
        return -1;
    }

    device->spec = *obtained;

    if (InitDevice(device) < 0) {
        return -1;
    }

    *obtained = device->spec;

    device->is_spec_set = SDL_TRUE;

    return 0;
}

int
SDL_VideoCaptureAcquireFrame(SDL_VideoCaptureDeviceID id, SDL_VideoCaptureFrame *frame, Uint64 *ticks)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    Uint64 _ticks;

    if (!device) {
        return SDL_InvalidParamError("id");
    }

    if (!frame) {
        return SDL_InvalidParamError("frame");
    }

    SDL_zerop(frame);
    frame->fd = -1;
    frame->clientbuffer = NULL;

    if (ticks) {
        *ticks = 0;
    } else {
        ticks = &_ticks;
        _ticks = 0;
    }

    if (device->thread == NULL) {
        int ret;

        /* wait a frame */
        while ((ret = AcquireFrame(device, frame, ticks)) == 0) {
            if (frame->num_planes) {
                return 0;
            }
        }
        return -1;
    } else {
        entry_t *entry = NULL;

        SDL_LockMutex(device->device_lock);
        SDL_ListPop(&device->buffer_queue, (void**)&entry);
        SDL_UnlockMutex(device->device_lock);

        if (entry) {
            *frame = entry->frame;
            *ticks = entry->ticks;
            SDL_free(entry);

            /* Error from thread */
            if (frame->num_planes == 0 && *ticks == 0) {
                SDL_SetError("error from acquisition thread");
                return -1;
            }


        } else {
            // Queue is empty. Not an error.
        }
    }

    return 0;
}

int
SDL_VideoCaptureReleaseFrame(SDL_VideoCaptureDeviceID id, SDL_VideoCaptureFrame *frame)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);

    if (!device) {
        return SDL_InvalidParamError("id");
    }

    if (frame == NULL) {
        return SDL_InvalidParamError("frame");
    }

    if (ReleaseFrame(device, frame) < 0) {
        return -1;
    }

    SDL_zerop(frame);

    return 0;
}


int
SDL_GetNumVideoCaptureFormats(SDL_VideoCaptureDeviceID id)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        return SDL_InvalidParamError("id");
    }
    return GetNumFormats(device);
}

int
SDL_GetVideoCaptureFormat(SDL_VideoCaptureDeviceID id, int index, Uint32 *format, Uint32 *type)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        return SDL_InvalidParamError("id");
    }
    if (!format) {
        return SDL_InvalidParamError("format");
    }
    if (!type) {
        return SDL_InvalidParamError("type");
    }
    *format = 0;
    *type = 0;
    return GetFormat(device, index, format, type);
}

int
SDL_GetNumVideoCaptureFrameSizes(SDL_VideoCaptureDeviceID id, Uint32 format)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        return SDL_InvalidParamError("id");
    }
    return GetNumFrameSizes(device, format);
}

int
SDL_GetVideoCaptureFrameSize(SDL_VideoCaptureDeviceID id, Uint32 format, int index, Uint32 *width, Uint32 *height)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);
    if (!device) {
        return SDL_InvalidParamError("id");
    }
    if (!width) {
        return SDL_InvalidParamError("width");
    }
    if (!height) {
        return SDL_InvalidParamError("height");
    }
    *width = 0;
    *height = 0;
    return GetFrameSize(device, format, index, width, height);
}

SDL_VideoCaptureDeviceID
SDL_OpenVideoCaptureWithSpec(
        const char *device_name,
        const SDL_VideoCaptureSpec *desired,
        SDL_VideoCaptureSpec *obtained,
        int allowed_changes)
{
    SDL_VideoCaptureDeviceID id;

    if ((id = SDL_OpenVideoCapture(device_name)) == SDL_VIDEO_CAPTURE_INVALID) {
        return SDL_VIDEO_CAPTURE_INVALID;
    }

    if (SDL_SetVideoCaptureSpec(id, desired, obtained, allowed_changes) < 0) {
        SDL_CloseVideoCapture(id);
        return SDL_VIDEO_CAPTURE_INVALID;
    }
    return id;
}

SDL_VideoCaptureStatus
SDL_GetVideoCaptureStatus(SDL_VideoCaptureDeviceID id)
{
    SDL_VideoCaptureDevice *device = get_video_capture_device(id);

    if (device == NULL) {
        return SDL_VIDEO_CAPTURE_INIT;
    }

    if (device->is_spec_set == SDL_FALSE) {
        return SDL_VIDEO_CAPTURE_INIT;
    }

    if (SDL_AtomicGet(&device->shutdown)) {
        return SDL_VIDEO_CAPTURE_STOPPED;
    }

    if (SDL_AtomicGet(&device->enabled)) {
        return SDL_VIDEO_CAPTURE_PLAYING;
    }
    return SDL_VIDEO_CAPTURE_INIT;
}

int
SDL_VideoCaptureInit(void)
{
    SDL_zeroa(open_devices);
    return 0;
}

void
SDL_QuitVideoCapture(void)
{
    Uint32 i;
    for (i = 0; i < SDL_arraysize(open_devices); i++) {
        close_device(open_devices[i]);
    }

    SDL_zeroa(open_devices);
}


/* ---------------------------
 *
 * DRIVER CODE using V4L2
 */

#if defined(__linux__) && !defined(__ANDROID__)

enum io_method {
    IO_METHOD_READ,
    IO_METHOD_MMAP,
    IO_METHOD_USERPTR
};

struct buffer {
    void   *start;
    size_t  length;
    int dmabuf_fd;
    int available; /* Is available in userspace */
};

struct SDL_PrivateVideoCaptureData
{
    int fd;
    enum io_method io;
    Uint32 nb_buffers;
    struct buffer *buffers;
    int first_start;
    int driver_pitch;
};

#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>              /* low-level i/o */
#include <errno.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <linux/videodev2.h>

static int
xioctl(int fh, int request, void *arg)
{
    int r;

    do {
        r = ioctl(fh, request, arg);
    } while (r == -1 && errno == EINTR);

    return r;
}

/* -1:error  1:frame 0:no frame*/
static int
acquire_frame(_THIS, SDL_VideoCaptureFrame *frame)
{
    struct v4l2_buffer buf;
    unsigned int i;

    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;
    int size = (int) _this->hidden->buffers[0].length;

    switch (io) {
        case IO_METHOD_READ:
            if (read(fd, _this->hidden->buffers[0].start, size) == -1) {
                switch (errno) {
                    case EAGAIN:
                        return 0;

                    case EIO:
                        /* Could ignore EIO, see spec. */

                        /* fall through */

                    default:
                        return SDL_SetError("read");
                }
            }

            frame->num_planes = 1;
            frame->data[0] = _this->hidden->buffers[0].start;
            frame->row_stride[0] = _this->hidden->driver_pitch;
            break;

        case IO_METHOD_MMAP:
            SDL_zero(buf);

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_MMAP;

            if (xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
                switch (errno) {
                    case EAGAIN:
                        return 0;

                    case EIO:
                        /* Could ignore EIO, see spec. */

                        /* fall through */

                    default:
                        return SDL_SetError("VIDIOC_DQBUF: %d", errno);
                }
            }

            if (buf.index >= _this->hidden->nb_buffers) {
                return SDL_SetError("invalid buffer index");
            }

            frame->num_planes = 1;
            frame->data[0] = _this->hidden->buffers[buf.index].start;
            frame->row_stride[0] = _this->hidden->driver_pitch;
            frame->fd = _this->hidden->buffers[buf.index].dmabuf_fd;
            _this->hidden->buffers[buf.index].available = 1;

#if DEBUG_VIDEO_CAPTURE_CAPTURE
            SDL_Log("debug mmap: image %d/%d  num_planes:%d data[0]=%p", buf.index, _this->hidden->nb_buffers, frame->num_planes, frame->data[0]);
#endif
            break;

        case IO_METHOD_USERPTR:
            SDL_zero(buf);

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_USERPTR;

            if (xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
                switch (errno) {
                    case EAGAIN:
                        return 0;

                    case EIO:
                        /* Could ignore EIO, see spec. */

                        /* fall through */

                    default:
                        return SDL_SetError("VIDIOC_DQBUF");
                }
            }

            for (i = 0; i < _this->hidden->nb_buffers; ++i) {
                if (buf.m.userptr == (unsigned long)_this->hidden->buffers[i].start && buf.length == size) {
                    break;
                }
            }

            if (i >= _this->hidden->nb_buffers) {
                return SDL_SetError("invalid buffer index");
            }

            frame->num_planes = 1;
            frame->data[0] = (void*)buf.m.userptr;
            frame->row_stride[0] = _this->hidden->driver_pitch;
            _this->hidden->buffers[i].available = 1;
#if DEBUG_VIDEO_CAPTURE_CAPTURE
            SDL_Log("debug userptr: image %d/%d  num_planes:%d data[0]=%p", buf.index, _this->hidden->nb_buffers, frame->num_planes, frame->data[0]);
#endif
            break;
    }

    return 1;
}


int
ReleaseFrame(_THIS, SDL_VideoCaptureFrame *frame)
{
    struct v4l2_buffer buf;
    unsigned int i;
    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;

    for (i = 0; i < _this->hidden->nb_buffers; ++i) {
        if (frame->num_planes && frame->data[0] == _this->hidden->buffers[i].start) {
            break;
        }
    }

    if (i >= _this->hidden->nb_buffers) {
        return SDL_SetError("invalid buffer index");
    }

    switch (io) {
        case IO_METHOD_READ:
            break;

        case IO_METHOD_MMAP:
            SDL_zero(buf);

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_MMAP;
            buf.index = i;

            if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
                return SDL_SetError("VIDIOC_QBUF");
            }
            _this->hidden->buffers[i].available = 0;
            break;

        case IO_METHOD_USERPTR:
            SDL_zero(buf);

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_USERPTR;
            buf.index = i;
            buf.m.userptr = (unsigned long)frame->data[0];
            buf.length = (int) _this->hidden->buffers[i].length;

            if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
                return SDL_SetError("VIDIOC_QBUF");
            }
            _this->hidden->buffers[i].available = 0;
            break;
    }

    return 0;
}


int
AcquireFrame(_THIS, SDL_VideoCaptureFrame *frame, Uint64 *ticks)
{
    fd_set fds;
    struct timeval tv;
    int ret;

    int fd = _this->hidden->fd;

    FD_ZERO(&fds);
    FD_SET(fd, &fds);

    /* Timeout. */
    tv.tv_sec = 0;
    tv.tv_usec = 300 * 1000;

    ret = select(fd + 1, &fds, NULL, NULL, &tv);

    if (ret == -1) {
        if (errno == EINTR) {
#if DEBUG_VIDEO_CAPTURE_CAPTURE
            SDL_Log("continue ..");
#endif
            return 0;
        }
        return SDL_SetError("select");
    }

    if (ret == 0) {
        /* Timeout. Not an error */
        SDL_SetError("timeout select");
        return 0;
    }

    ret = acquire_frame(_this, frame);
    if (ret < 0) {
        return -1;
    }

    if (ret == 1){
        if (ticks) {
            *ticks = SDL_GetTicks();
        }
    } else if (ret == 0) {
#if DEBUG_VIDEO_CAPTURE_CAPTURE
        SDL_Log("No frame continue: %s", SDL_GetError());
#endif
    }

    /* EAGAIN - continue select loop. */
    return 0;
}


int
StopCapture(_THIS)
{
    enum v4l2_buf_type type;
    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;

    switch (io) {
        case IO_METHOD_READ:
            break;

        case IO_METHOD_MMAP:
        case IO_METHOD_USERPTR:
            type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            if (xioctl(fd, VIDIOC_STREAMOFF, &type) == -1) {
                return SDL_SetError("VIDIOC_STREAMOFF");
            }
            break;
    }

    return 0;
}

static int
enqueue_buffers(_THIS)
{
    unsigned int i;
    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;
    switch (io) {
        case IO_METHOD_READ:
            break;

        case IO_METHOD_MMAP:
            for (i = 0; i < _this->hidden->nb_buffers; ++i) {
                if (_this->hidden->buffers[i].available == 0) {
                    struct v4l2_buffer buf;

                    SDL_zero(buf);
                    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                    buf.memory = V4L2_MEMORY_MMAP;
                    buf.index = i;

                    if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
                        return SDL_SetError("VIDIOC_QBUF");
                    }
                }
            }
            break;

        case IO_METHOD_USERPTR:
            for (i = 0; i < _this->hidden->nb_buffers; ++i) {
                if (_this->hidden->buffers[i].available == 0) {
                    struct v4l2_buffer buf;

                    SDL_zero(buf);
                    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                    buf.memory = V4L2_MEMORY_USERPTR;
                    buf.index = i;
                    buf.m.userptr = (unsigned long)_this->hidden->buffers[i].start;
                    buf.length = (int) _this->hidden->buffers[i].length;

                    if (xioctl(fd, VIDIOC_QBUF, &buf) == -1) {
                        return SDL_SetError("VIDIOC_QBUF");
                    }
                }
            }
            break;
    }
    return 0;
}

static int
pre_enqueue_buffers(_THIS)
{
    struct v4l2_requestbuffers req;
    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;

    switch (io) {
        case IO_METHOD_READ:
            break;

        case IO_METHOD_MMAP:
            {
                SDL_zero(req);
                req.count = _this->hidden->nb_buffers;
                req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                req.memory = V4L2_MEMORY_MMAP;

                if (xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
                    if (errno == EINVAL) {
                        return SDL_SetError("Does not support memory mapping");
                    } else {
                        return SDL_SetError("VIDIOC_REQBUFS");
                    }
                }

                if (req.count < 2) {
                    return SDL_SetError("Insufficient buffer memory");
                }

                _this->hidden->nb_buffers = req.count;
            }
            break;

        case IO_METHOD_USERPTR:
            {
                SDL_zero(req);
                req.count  = _this->hidden->nb_buffers;
                req.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                req.memory = V4L2_MEMORY_USERPTR;

                if (xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
                    if (errno == EINVAL) {
                        return SDL_SetError("Does not support user pointer i/o");
                    } else {
                        return SDL_SetError("VIDIOC_REQBUFS");
                    }
                }
            }
            break;
    }
    return 0;
}

int
StartCapture(_THIS)
{
    enum v4l2_buf_type type;

    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;


    if (_this->hidden->first_start == 0) {
        _this->hidden->first_start = 1;
    } else {
        Uint32 old = _this->hidden->nb_buffers;
        // TODO mmap; doesn't work with stop->start
#if 1
        /* Can change nb_buffers for mmap */
        if (pre_enqueue_buffers(_this) < 0) {
            return -1;
        }
        if (old != _this->hidden->nb_buffers) {
            SDL_SetError("different nb of buffers requested");
            return -1;
        }
#endif
        _this->hidden->first_start = 1;
    }

    if (enqueue_buffers(_this) < 0) {
        return -1;
    }

    switch (io) {
        case IO_METHOD_READ:
            break;

        case IO_METHOD_MMAP:
        case IO_METHOD_USERPTR:
            type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            if (xioctl(fd, VIDIOC_STREAMON, &type) == -1) {
                return SDL_SetError("VIDIOC_STREAMON");
            }
            break;
    }

    return 0;
}

static int alloc_buffer_read(_THIS, unsigned int buffer_size)
{
    _this->hidden->buffers[0].length = buffer_size;
    _this->hidden->buffers[0].start = SDL_calloc(1, buffer_size);

    if (!_this->hidden->buffers[0].start) {
        return SDL_OutOfMemory();
    }
    return 0;
}

static int
alloc_buffer_mmap(_THIS)
{
    int fd = _this->hidden->fd;
    Uint32 i;
    for (i = 0; i < _this->hidden->nb_buffers; ++i) {
        struct v4l2_buffer buf;

        SDL_zero(buf);

        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = i;

        if (xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1) {
            return SDL_SetError("VIDIOC_QUERYBUF");
        }

        _this->hidden->buffers[i].length = buf.length;
        _this->hidden->buffers[i].start =
            mmap(NULL /* start anywhere */,
                    buf.length,
                    PROT_READ | PROT_WRITE /* required */,
                    MAP_SHARED /* recommended */,
                    fd, buf.m.offset);

        if (MAP_FAILED == _this->hidden->buffers[i].start) {
            return SDL_SetError("mmap");
        }
    }
    return 0;
}

static int
alloc_buffer_userp(_THIS, unsigned int buffer_size)
{
    int i;
    for (i = 0; i < _this->hidden->nb_buffers; ++i) {
        _this->hidden->buffers[i].length = buffer_size;
        _this->hidden->buffers[i].start = SDL_calloc(1, buffer_size);

        if (!_this->hidden->buffers[i].start) {
            return SDL_OutOfMemory();
        }
    }
    return 0;
}

static int
export_dmabuf(_THIS)
{
    int fd = _this->hidden->fd;
    Uint32 i;
    for (i = 0; i < _this->hidden->nb_buffers; ++i) {
        struct v4l2_exportbuffer expbuf;
        SDL_zero(expbuf);
        expbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        expbuf.index = i;
        expbuf.flags = O_RDWR;
        if (ioctl(fd, VIDIOC_EXPBUF, &expbuf) == -1) {
            SDL_SetError("VIDIOC_EXPBUF");
            return -1;
        }
        _this->hidden->buffers[i].dmabuf_fd = expbuf.fd;
    }
    return 0;
}

static Uint32
format_v4l2_2_sdl(Uint32 fmt)
{
    switch (fmt) {
#define CASE(x, y)  case x: return y
        CASE(V4L2_PIX_FMT_YUYV, SDL_PIXELFORMAT_YUY2);
        CASE(V4L2_PIX_FMT_MJPEG, SDL_PIXELFORMAT_UNKNOWN);
#undef CASE
        default:
            SDL_Log("Unknown format V4L2_PIX_FORMAT '%d'", fmt);
            return SDL_PIXELFORMAT_UNKNOWN;
    }
}

static Uint32
format_sdl_2_v4l2(Uint32 fmt)
{
    switch (fmt) {
#define CASE(y, x)  case x: return y
        CASE(V4L2_PIX_FMT_YUYV, SDL_PIXELFORMAT_YUY2);
        CASE(V4L2_PIX_FMT_MJPEG, SDL_PIXELFORMAT_UNKNOWN);
#undef CASE
        default:
            return 0;
    }
}

int
GetNumFormats(_THIS)
{
    int fd = _this->hidden->fd;
    Uint32 i = 0;
    struct v4l2_fmtdesc fmtdesc;

    SDL_zero(fmtdesc);
    fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    while (ioctl(fd,VIDIOC_ENUM_FMT,&fmtdesc) == 0) {
        fmtdesc.index++;
        i++;
    }
    return i;
}

int
GetFormat(_THIS, int index, Uint32 *format, Uint32 *type)
{
    int fd = _this->hidden->fd;
    struct v4l2_fmtdesc fmtdesc;

    SDL_zero(fmtdesc);
    fmtdesc.index = index;
    fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(fd,VIDIOC_ENUM_FMT,&fmtdesc) == 0) {
        *format = format_v4l2_2_sdl(fmtdesc.pixelformat);
        if (fmtdesc.flags & V4L2_FMT_FLAG_EMULATED) {
            *type |= SDL_VIDEO_CAPTURE_TYPE_EMULATED;
        }
        if (fmtdesc.flags & V4L2_FMT_FLAG_COMPRESSED) {
            *type |= SDL_VIDEO_CAPTURE_TYPE_EMULATED;
        }
        return 0;
    }

    return -1;
}

int
GetNumFrameSizes(_THIS, Uint32 format)
{
    int fd = _this->hidden->fd;
    Uint32 i = 0;
    struct v4l2_frmsizeenum frmsizeenum;

    SDL_zero(frmsizeenum);
    frmsizeenum.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    frmsizeenum.pixel_format = format_sdl_2_v4l2(format);
    while (ioctl(fd,VIDIOC_ENUM_FRAMESIZES, &frmsizeenum) == 0) {
        frmsizeenum.index++;
        if (frmsizeenum.type == V4L2_FRMSIZE_TYPE_DISCRETE) {
            i++;
        } else if (frmsizeenum.type == V4L2_FRMSIZE_TYPE_STEPWISE) {
            i += (1 + (frmsizeenum.stepwise.max_width - frmsizeenum.stepwise.min_width) / frmsizeenum.stepwise.step_width)
                * (1 + (frmsizeenum.stepwise.max_height - frmsizeenum.stepwise.min_height) / frmsizeenum.stepwise.step_height);
        } else if (frmsizeenum.type == V4L2_FRMSIZE_TYPE_CONTINUOUS) {
            SDL_SetError("V4L2_FRMSIZE_TYPE_CONTINUOUS not handled");
        }
    }
    return i;
}

int
GetFrameSize(_THIS, Uint32 format, int index, Uint32 *width, Uint32 *height)
{
    int fd = _this->hidden->fd;
    struct v4l2_frmsizeenum frmsizeenum;
    int i = 0;

    SDL_zero(frmsizeenum);
    frmsizeenum.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    frmsizeenum.pixel_format = format_sdl_2_v4l2(format);
    while (ioctl(fd,VIDIOC_ENUM_FRAMESIZES, &frmsizeenum) == 0) {
        frmsizeenum.index++;

        if (frmsizeenum.type == V4L2_FRMSIZE_TYPE_DISCRETE) {
            if (i == index) {
                *width = frmsizeenum.discrete.width;
                *height = frmsizeenum.discrete.height;
                return 0;
            }
            i++;
        } else if (frmsizeenum.type == V4L2_FRMSIZE_TYPE_STEPWISE) {
            Uint32 w;
            for (w = frmsizeenum.stepwise.min_width; w <= frmsizeenum.stepwise.max_width; w += frmsizeenum.stepwise.step_width) {
                Uint32 h;
                for (h = frmsizeenum.stepwise.min_height; h <= frmsizeenum.stepwise.max_height; h += frmsizeenum.stepwise.step_height) {
                    if (i == index) {
                        *width = h;
                        *height = w;
                        return 0;
                    }
                    i++;
                }
            }
        } else if (frmsizeenum.type == V4L2_FRMSIZE_TYPE_CONTINUOUS) {
        }
    }

    return -1;
}



static void
dbg_v4l2_pixelformat(const char *str, int f) {
    SDL_Log("%s  V4L2_format=%d  %c%c%c%c", str, f,
                (f >> 0) & 0xff,
                (f >> 8) & 0xff,
                (f >> 16) & 0xff,
                (f >> 24) & 0xff);
}

int
GetDeviceSpec(_THIS, SDL_VideoCaptureSpec *spec)
{
    struct v4l2_format fmt;
    int fd = _this->hidden->fd;
    Uint32 min;

    SDL_zero(fmt);
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    /* Preserve original settings as set by v4l2-ctl for example */
    if (xioctl(fd, VIDIOC_G_FMT, &fmt) == -1) {
        return SDL_SetError("Error VIDIOC_G_FMT");
    }

    /* Buggy driver paranoia. */
    min = fmt.fmt.pix.width * 2;
    if (fmt.fmt.pix.bytesperline < min) {
        fmt.fmt.pix.bytesperline = min;
    }
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min) {
        fmt.fmt.pix.sizeimage = min;
    }

    //spec->width = fmt.fmt.pix.width;
    //spec->height = fmt.fmt.pix.height;
    _this->hidden->driver_pitch = fmt.fmt.pix.bytesperline;
    //spec->format = format_v4l2_2_sdl(fmt.fmt.pix.pixelformat);

    return 0;
}

int
InitDevice(_THIS)
{
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;

    int fd = _this->hidden->fd;
    enum io_method io = _this->hidden->io;
    int ret = -1;

    /* Select video input, video standard and tune here. */
    SDL_zero(cropcap);

    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0) {
        crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        crop.c = cropcap.defrect; /* reset to default */

        if (xioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
            switch (errno) {
                case EINVAL:
                    /* Cropping not supported. */
                    break;
                default:
                    /* Errors ignored. */
                    break;
            }
        }
    } else {
        /* Errors ignored. */
    }


    {
        struct v4l2_format fmt;
        SDL_zero(fmt);

        fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix.width       = _this->spec.width;
        fmt.fmt.pix.height      = _this->spec.height;


        fmt.fmt.pix.pixelformat = format_sdl_2_v4l2(_this->spec.format);
        //    fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
        fmt.fmt.pix.field       = V4L2_FIELD_ANY;

#if DEBUG_VIDEO_CAPTURE_CAPTURE
        SDL_Log("set SDL format %s", SDL_GetPixelFormatName(_this->spec.format));
        dbg_v4l2_pixelformat("set format", fmt.fmt.pix.pixelformat);
#endif

        if (xioctl(fd, VIDIOC_S_FMT, &fmt) == -1) {
            return SDL_SetError("Error VIDIOC_S_FMT");
        }
    }

    GetDeviceSpec(_this, &_this->spec);

    if (pre_enqueue_buffers(_this) < 0) {
        return -1;
    }

    {
        Uint32 i;
        _this->hidden->buffers = SDL_calloc(_this->hidden->nb_buffers, sizeof(*_this->hidden->buffers));
        if (!_this->hidden->buffers) {
            return SDL_OutOfMemory();
        }

        for (i = 0; i < _this->hidden->nb_buffers; ++i) {
            _this->hidden->buffers[i].dmabuf_fd = -1;
        }
    }

    {
        Uint32 sz, pitch;
        size_buffer(_this->spec.width, _this->spec.height, _this->spec.format, &sz, &pitch);

        switch (io) {
            case IO_METHOD_READ:
                ret = alloc_buffer_read(_this, sz);
                break;

            case IO_METHOD_MMAP:
                ret = alloc_buffer_mmap(_this);

                export_dmabuf(_this);
                break;

            case IO_METHOD_USERPTR:
                ret = alloc_buffer_userp(_this, sz);
                break;
        }
    }

    if (ret < 0) {
        return -1;
    }

    return 0;
}

void
CloseDevice(_THIS)
{
    if (!_this) {
        return;
    }

    if (_this->hidden) {
        if (_this->hidden->buffers) {
            unsigned int i;
            enum io_method io = _this->hidden->io;

            switch (io) {
                case IO_METHOD_READ:
                    SDL_free(_this->hidden->buffers[0].start);
                    break;

                case IO_METHOD_MMAP:
                    for (i = 0; i < _this->hidden->nb_buffers; ++i) {
                        if (_this->hidden->buffers[i].dmabuf_fd != -1) {
                            close(_this->hidden->buffers[i].dmabuf_fd);
                        }
                        if (munmap(_this->hidden->buffers[i].start, _this->hidden->buffers[i].length) == -1) {
                            SDL_SetError("munmap");
                        }
                    }
                    break;

                case IO_METHOD_USERPTR:
                    for (i = 0; i < _this->hidden->nb_buffers; ++i) {
                        SDL_free(_this->hidden->buffers[i].start);
                    }
                    break;
            }

            SDL_free(_this->hidden->buffers);
        }

        if (_this->hidden->fd != -1) {
            if (close(_this->hidden->fd)) {
                SDL_SetError("close video capture device");
            }
        }
        SDL_free(_this->hidden);

        _this->hidden = NULL;
    }
}


int
OpenDevice(_THIS)
{
    struct stat st;
    struct v4l2_capability cap;
    int fd;
    enum io_method io;

    _this->hidden = (struct SDL_PrivateVideoCaptureData *) SDL_calloc(1, sizeof (struct SDL_PrivateVideoCaptureData));
    if (_this->hidden == NULL) {
        SDL_OutOfMemory();
        return -1;
    }

    _this->hidden->fd = -1;

    if (stat(_this->dev_name, &st) == -1) {
        SDL_SetError("Cannot identify '%s': %d, %s", _this->dev_name, errno, strerror(errno));
        return -1;
    }

    if (!S_ISCHR(st.st_mode)) {
        SDL_SetError("%s is no device", _this->dev_name);
        return -1;
    }

    fd = open(_this->dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
    if (fd == -1) {
        SDL_SetError("Cannot open '%s': %d, %s", _this->dev_name, errno, strerror(errno));
        return -1;
    }

    _this->hidden->fd = fd;
    _this->hidden->io = IO_METHOD_MMAP;
//    _this->hidden->io = IO_METHOD_USERPTR;
//    _this->hidden->io = IO_METHOD_READ;
//
    if (_this->hidden->io == IO_METHOD_READ) {
        _this->hidden->nb_buffers = 1;
    } else {
        _this->hidden->nb_buffers = 8; /* Number of image as internal buffer, */
    }
    io = _this->hidden->io;

    if (xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
        if (errno == EINVAL) {
            return SDL_SetError("%s is no V4L2 device", _this->dev_name);
        } else {
            return SDL_SetError("Error VIDIOC_QUERYCAP errno=%d device%s is no V4L2 device", errno, _this->dev_name);
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        return SDL_SetError("%s is no video capture device", _this->dev_name);
    }

#if 0
    if ((cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        SDL_Log("%s is video capture device - single plane", _this->dev_name);
    }
    if ((cap.capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE)) {
        SDL_Log("%s is video capture device - multiple planes", _this->dev_name);
    }
#endif

    switch (io) {
        case IO_METHOD_READ:
            if (!(cap.capabilities & V4L2_CAP_READWRITE)) {
                return SDL_SetError("%s does not support read i/o", _this->dev_name);
            }
            break;

        case IO_METHOD_MMAP:
        case IO_METHOD_USERPTR:
            if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
                return SDL_SetError("%s does not support streaming i/o", _this->dev_name);
            }
            break;
    }




    return 0;
}



int
GetDeviceName(int index, char *buf, int size) {
    SDL_snprintf(buf, size, "/dev/video%d", index);
    return 0;
}

int
GetNumDevices(void) {
    int num;
    for (num = 0; num < 128; num++) {
        const char *filename = SDL_GetVideoCaptureDeviceName(num);
        SDL_RWops *src = SDL_RWFromFile(filename, "rb");
        if (src == NULL) {
            // When file does not exist, an error is set. Clear it.
            SDL_ClearError();
            return num;
        }
        SDL_RWclose(src);
    }
    return num;
}

#elif defined(__ANDROID__) && __ANDROID_API__ >= 24

/*
 * APP_PLATFORM=android-24
 * minSdkVersion=24
 *
 * link with: -lcamera2ndk -lmediandk
 *
 * AndroidManifest.xml:
 *   <uses-permission android:name="android.permission.CAMERA"></uses-permission>
 *   <uses-feature android:name="android.hardware.camera" />
 *
 * for DMA BUF:
 * API 26
 *      AImage_getHardwareBuffer
 *      eglGetNativeClientBufferANDROID
 *
 *
 */

#include <camera/NdkCameraDevice.h>
#include <camera/NdkCameraManager.h>
#include <media/NdkImage.h>
#include <media/NdkImageReader.h>
#include "SDL_egl.h"

#include "../core/android/SDL_android.h"


static ACameraManager *cameraMgr = NULL;
static ACameraIdList *cameraIdList = NULL;

typedef EGLClientBuffer (*eglGetNativeClientBufferANDROID_t) (const struct AHardwareBuffer *buffer);
static eglGetNativeClientBufferANDROID_t eglGetNativeClientBufferANDROID = NULL;


static void
create_cameraMgr(void)
{
    if (cameraMgr == NULL) {

        if (!Android_JNI_RequestPermission("android.permission.CAMERA")) {
            SDL_SetError("This app doesn't have CAMERA permission");
            return;
        }

        cameraMgr = ACameraManager_create();
        if (cameraMgr == NULL) {
            SDL_Log("Error creating ACameraManager");
        } else {
            SDL_Log("Create ACameraManager");
        }

        eglGetNativeClientBufferANDROID = (eglGetNativeClientBufferANDROID_t) SDL_GL_GetProcAddress("eglGetNativeClientBufferANDROID");
        if (eglGetNativeClientBufferANDROID == NULL) {
            SDL_Log("Cannot load function: eglGetNativeClientBufferANDROID");
        }
    }
}

static void
delete_cameraMgr(void)
{
    if (cameraIdList) {
        ACameraManager_deleteCameraIdList(cameraIdList);
        cameraIdList = NULL;
    }

    if (cameraMgr) {
        ACameraManager_delete(cameraMgr);
        cameraMgr = NULL;
    }
}

struct SDL_PrivateVideoCaptureData
{
    ACameraDevice *device;
    ACameraCaptureSession *session;
    ACameraDevice_StateCallbacks dev_callbacks;
    ACameraCaptureSession_stateCallbacks capture_callbacks;
    ACaptureSessionOutputContainer *sessionOutputContainer;
    AImageReader *reader;
    int num_formats;
    int count_formats[6]; // see format_2_id
};


/**/
#define FORMAT_SDL SDL_PIXELFORMAT_NV12

static int
format_2_id(int fmt) {
     switch (fmt) {
#define CASE(x, y)  case x: return y
        CASE(FORMAT_SDL, 0);
        CASE(SDL_PIXELFORMAT_RGB565, 1);
        CASE(SDL_PIXELFORMAT_RGB888, 2);
        CASE(SDL_PIXELFORMAT_RGBA8888, 3);
        CASE(SDL_PIXELFORMAT_RGBX8888, 4);
        CASE(SDL_PIXELFORMAT_UNKNOWN, 5);
#undef CASE
        default:
                return 5;
    }
}

static int
id_2_format(int fmt) {
     switch (fmt) {
#define CASE(x, y)  case y: return x
        CASE(FORMAT_SDL, 0);
        CASE(SDL_PIXELFORMAT_RGB565, 1);
        CASE(SDL_PIXELFORMAT_RGB888, 2);
        CASE(SDL_PIXELFORMAT_RGBA8888, 3);
        CASE(SDL_PIXELFORMAT_RGBX8888, 4);
        CASE(SDL_PIXELFORMAT_UNKNOWN, 5);
#undef CASE
        default:
            return SDL_PIXELFORMAT_UNKNOWN;
    }
}

static Uint32
format_android_2_sdl(Uint32 fmt)
{
    switch (fmt) {
#define CASE(x, y)  case x: return y
        CASE(AIMAGE_FORMAT_YUV_420_888, FORMAT_SDL);
        CASE(AIMAGE_FORMAT_RGB_565,     SDL_PIXELFORMAT_RGB565);
        CASE(AIMAGE_FORMAT_RGB_888,     SDL_PIXELFORMAT_RGB888);
        CASE(AIMAGE_FORMAT_RGBA_8888,   SDL_PIXELFORMAT_RGBA8888);
        CASE(AIMAGE_FORMAT_RGBX_8888,   SDL_PIXELFORMAT_RGBX8888);

        CASE(AIMAGE_FORMAT_RGBA_FP16,   SDL_PIXELFORMAT_UNKNOWN); // 64bits
        CASE(AIMAGE_FORMAT_RAW_PRIVATE, SDL_PIXELFORMAT_UNKNOWN);
        CASE(AIMAGE_FORMAT_JPEG,        SDL_PIXELFORMAT_UNKNOWN);
#undef CASE
        default:
            SDL_Log("Unknown format AIMAGE_FORMAT '%d'", fmt);
            return SDL_PIXELFORMAT_UNKNOWN;
    }
}

static Uint32
format_sdl_2_android(Uint32 fmt)
{
    switch (fmt) {
#define CASE(x, y)  case y: return x
        CASE(AIMAGE_FORMAT_YUV_420_888, FORMAT_SDL);
        CASE(AIMAGE_FORMAT_RGB_565,     SDL_PIXELFORMAT_RGB565);
        CASE(AIMAGE_FORMAT_RGB_888,     SDL_PIXELFORMAT_RGB888);
        CASE(AIMAGE_FORMAT_RGBA_8888,   SDL_PIXELFORMAT_RGBA8888);
        CASE(AIMAGE_FORMAT_RGBX_8888,   SDL_PIXELFORMAT_RGBX8888);
#undef CASE
        default:
            return 0;
    }
}


static void
onDisconnected(void *context, ACameraDevice *device)
{
    // SDL_VideoCaptureDevice *_this = (SDL_VideoCaptureDevice *) context;
    SDL_Log("CB onDisconnected");
}

static void
onError(void *context, ACameraDevice *device, int error)
{
    // SDL_VideoCaptureDevice *_this = (SDL_VideoCaptureDevice *) context;
    SDL_Log("CB onError");
}


static void
onClosed(void* context, ACameraCaptureSession *session)
{
    // SDL_VideoCaptureDevice *_this = (SDL_VideoCaptureDevice *) context;
    SDL_Log("CB onClosed");
}

static void
onReady(void* context, ACameraCaptureSession *session)
{
    // SDL_VideoCaptureDevice *_this = (SDL_VideoCaptureDevice *) context;
    SDL_Log("CB onReady");
}

static void
onActive(void* context, ACameraCaptureSession *session)
{
    // SDL_VideoCaptureDevice *_this = (SDL_VideoCaptureDevice *) context;
    SDL_Log("CB onActive");
}

int
OpenDevice(_THIS)
{
    camera_status_t res;

    _this->hidden = (struct SDL_PrivateVideoCaptureData *) SDL_calloc(1, sizeof (struct SDL_PrivateVideoCaptureData));
    if (_this->hidden == NULL) {
        SDL_OutOfMemory();
        return -1;
    }

    create_cameraMgr();

    _this->hidden->dev_callbacks.context = (void *) _this;
    _this->hidden->dev_callbacks.onDisconnected = onDisconnected;
    _this->hidden->dev_callbacks.onError = onError;

    res = ACameraManager_openCamera(cameraMgr, _this->dev_name, &_this->hidden->dev_callbacks, &_this->hidden->device);
    if (res != ACAMERA_OK) {
        goto error;
    }

    return 0;

error:
    return -1;
}

void
CloseDevice(_THIS)
{
    if (_this && _this->hidden) {
        if (_this->hidden->session) {
            ACameraCaptureSession_close(_this->hidden->session);
        }

        if (_this->hidden->sessionOutputContainer) {
            ACaptureSessionOutputContainer_free(_this->hidden->sessionOutputContainer);
        }

        if (_this->hidden->reader) {
            AImageReader_delete(_this->hidden->reader);
        }

        if (_this->hidden->device) {
            ACameraDevice_close(_this->hidden->device);
        }

        SDL_free(_this->hidden);

        _this->hidden = NULL;
    }


    {
        Uint32 i;
        int all_closed = 1;
        for (i = 0; i < SDL_arraysize(open_devices); i++) {
            if (open_devices[i]) {
                all_closed = 0;
                break;
            }
        }

        if (all_closed) {
            delete_cameraMgr();
        }
    }

}

int
InitDevice(_THIS)
{
    Uint32 sz, p;
    size_buffer(_this->spec.width, _this->spec.height, _this->spec.format, &sz, &p);
    SDL_Log("size_buffer: %d x %d", _this->spec.width, _this->spec.height);
    return 0;
}

int
GetDeviceSpec(_THIS, SDL_VideoCaptureSpec *spec)
{
    if (spec) {
        *spec = _this->spec;
        return 0;
    }
    return -1;
}

int
StartCapture(_THIS)
{
    camera_status_t res;
    media_status_t res2;
    ANativeWindow *window = NULL;

    res2 = AImageReader_new(_this->spec.width, _this->spec.height, format_sdl_2_android(_this->spec.format), 10 /* nb buffers */, &_this->hidden->reader);
    if (res2 != AMEDIA_OK) {
        SDL_SetError("Error AImageReader_new");
        goto error;
    }
    res2 = AImageReader_getWindow(_this->hidden->reader, &window);
    if (res2 != AMEDIA_OK) {
        SDL_SetError("Error AImageReader_new");
        goto error;

    }

    ACaptureSessionOutput *sessionOutput;
    res = ACaptureSessionOutput_create(window, &sessionOutput);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACaptureSessionOutput_create");
        goto error;
    }
    res = ACaptureSessionOutputContainer_create(&_this->hidden->sessionOutputContainer);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACaptureSessionOutputContainer_create");
        goto error;
    }
    res = ACaptureSessionOutputContainer_add(_this->hidden->sessionOutputContainer, sessionOutput);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACaptureSessionOutputContainer_add");
        goto error;
    }




    ACameraOutputTarget *outputTarget;
    res = ACameraOutputTarget_create(window, &outputTarget);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACameraOutputTarget_create");
        goto error;
    }


    ACaptureRequest *request;
    res = ACameraDevice_createCaptureRequest(_this->hidden->device, TEMPLATE_RECORD, &request);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACameraDevice_createCaptureRequest");
        goto error;
    }

    res = ACaptureRequest_addTarget(request, outputTarget);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACaptureRequest_addTarget");
        goto error;
    }


    _this->hidden->capture_callbacks.context = (void *) _this;
    _this->hidden->capture_callbacks.onClosed = onClosed;
    _this->hidden->capture_callbacks.onReady = onReady;
    _this->hidden->capture_callbacks.onActive = onActive;

    res = ACameraDevice_createCaptureSession(_this->hidden->device,
            _this->hidden->sessionOutputContainer,
            &_this->hidden->capture_callbacks,
            &_this->hidden->session);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACameraDevice_createCaptureSession");
        goto error;
    }

    res = ACameraCaptureSession_setRepeatingRequest(_this->hidden->session, NULL, 1, &request, NULL);
    if (res != ACAMERA_OK) {
        SDL_SetError("Error ACameraDevice_createCaptureSession");
        goto error;
    }

    return 0;

error:
    return -1;
}

int
StopCapture(_THIS)
{
    ACameraCaptureSession_close(_this->hidden->session);
    _this->hidden->session = NULL;
    return 0;
}

int
AcquireFrame(_THIS, SDL_VideoCaptureFrame *frame, Uint64 *ticks)
{
    media_status_t res;
    AImage *image;
    res = AImageReader_acquireNextImage(_this->hidden->reader, &image);
    if (res == AMEDIA_IMGREADER_NO_BUFFER_AVAILABLE ) {

        SDL_Delay(20); // TODO fix some delay
#if DEBUG_VIDEO_CAPTURE_CAPTURE
//        SDL_Log("AImageReader_acquireNextImage: AMEDIA_IMGREADER_NO_BUFFER_AVAILABLE");
#endif
        return 0;
    } else if (res == AMEDIA_OK ) {
        int i = 0;
        int32_t numPlanes = 0;
        AImage_getNumberOfPlanes(image, &numPlanes);

        *ticks = SDL_GetTicks64();

        for (i = 0; i < numPlanes && i < 3; i++) {
            int dataLength = 0;
            int rowStride = 0;
            uint8_t *data = NULL;
            frame->num_planes += 1;
            AImage_getPlaneRowStride(image, i, &rowStride);
            res = AImage_getPlaneData(image, i, &data, &dataLength);
            if (res == AMEDIA_OK) {
                frame->data[i] = data;
                frame->row_stride[i] = rowStride;
            }
        }

        if (frame->num_planes == 3) {
            /* plane 2 and 3 are interleaved NV12. SDL only takes two planes for this format */
            int pixelStride = 0;
            AImage_getPlanePixelStride(image, 1, &pixelStride);
            if (pixelStride == 2) {
                frame->num_planes -= 1;
            }
        }

#if __ANDROID_API__ >= 26
        {
            AHardwareBuffer *buffer = NULL;
            AImage_getHardwareBuffer(image, &buffer);
            if (buffer && eglGetNativeClientBufferANDROID) {
                frame->clientbuffer = eglGetNativeClientBufferANDROID(buffer);
            }
        }
#endif
        frame->internal = (void*)image;
        return 0;
    }

    SDL_SetError("Error AImageReader_acquireNextImage");
    return -1;
}

int
ReleaseFrame(_THIS, SDL_VideoCaptureFrame *frame)
{
    if (frame->internal){
        AImage_delete((AImage *)frame->internal);
    }
    return 0;
}

int
GetNumFormats(_THIS)
{
    camera_status_t res;
    int i;
    int unknown = 0;
    ACameraMetadata *metadata;
    ACameraMetadata_const_entry entry;

    if (_this->hidden->num_formats != 0) {
        return _this->hidden->num_formats;
    }

    res = ACameraManager_getCameraCharacteristics(cameraMgr, _this->dev_name, &metadata);
    if (res != ACAMERA_OK) {
        return -1;
    }

    res = ACameraMetadata_getConstEntry(metadata, ACAMERA_SCALER_AVAILABLE_STREAM_CONFIGURATIONS, &entry);
    if (res != ACAMERA_OK) {
        return -1;
    }

    SDL_Log("got entry ACAMERA_SCALER_AVAILABLE_STREAM_CONFIGURATIONS");

    for (i = 0; i < entry.count; i += 4) {
        int32_t format = entry.data.i32[i + 0];

        int32_t type = entry.data.i32[i + 3];
        Uint32 fmt;

        if (type == ACAMERA_SCALER_AVAILABLE_STREAM_CONFIGURATIONS_INPUT) {
            continue;
        }

        fmt = format_android_2_sdl(format);
        _this->hidden->count_formats[format_2_id(fmt)] += 1;

#if DEBUG_VIDEO_CAPTURE_CAPTURE
        if (fmt != SDL_PIXELFORMAT_UNKNOWN) {
            int32_t w = entry.data.i32[i + 1];
            int32_t h = entry.data.i32[i + 2];
            SDL_Log("Got format android 0x%08x -> %s %d x %d", format, SDL_GetPixelFormatName(fmt), w, h);
        } else {
            unknown += 1;
        }
#endif
    }

#if DEBUG_VIDEO_CAPTURE_CAPTURE
        if (unknown) {
            SDL_Log("Got unknown android");
        }
#endif


    if ( _this->hidden->count_formats[0]) _this->hidden->num_formats += 1;
    if ( _this->hidden->count_formats[1]) _this->hidden->num_formats += 1;
    if ( _this->hidden->count_formats[2]) _this->hidden->num_formats += 1;
    if ( _this->hidden->count_formats[3]) _this->hidden->num_formats += 1;
    if ( _this->hidden->count_formats[4]) _this->hidden->num_formats += 1;
    if ( _this->hidden->count_formats[5]) _this->hidden->num_formats += 1;

    return _this->hidden->num_formats;
}

int
GetFormat(_THIS, int index, Uint32 *format, Uint32 *type)
{
    int i;
    int i2 = 0;

    if (_this->hidden->num_formats == 0) {
        GetNumFormats(_this);
    }

    if (index < 0 || index >= _this->hidden->num_formats) {
        return -1;
    }

    for (i = 0; i < SDL_arraysize(_this->hidden->count_formats); i++) {
        if (_this->hidden->count_formats[i] == 0) {
            continue;
        }

        if (i2 == index) {
            *format = id_2_format(i);
            *type = 0;
        }

        i2++;

    }
    return 0;
}

int
GetNumFrameSizes(_THIS, Uint32 format)
{
    int i, i2 = 0, index;
    if (_this->hidden->num_formats == 0) {
        GetNumFormats(_this);
    }

    index = format_2_id(format);

    for (i = 0; i < SDL_arraysize(_this->hidden->count_formats); i++) {
        if (_this->hidden->count_formats[i] == 0) {
            continue;
        }

        if (i2 == index) {
            /* number of resolution for this format */
            return _this->hidden->count_formats[i];
        }

        i2++;
    }

    return -1;
}

int
GetFrameSize(_THIS, Uint32 format, int index, Uint32 *width, Uint32 *height)
{
    camera_status_t res;
    int i, i2 = 0;
    ACameraMetadata *metadata;
    ACameraMetadata_const_entry entry;

    if (_this->hidden->num_formats == 0) {
        GetNumFormats(_this);
    }

    res = ACameraManager_getCameraCharacteristics(cameraMgr, _this->dev_name, &metadata);
    if (res != ACAMERA_OK) {
        return -1;
    }

    res = ACameraMetadata_getConstEntry(metadata, ACAMERA_SCALER_AVAILABLE_STREAM_CONFIGURATIONS, &entry);
    if (res != ACAMERA_OK) {
        return -1;
    }

    for (i = 0; i < entry.count; i += 4) {
        int32_t f = entry.data.i32[i + 0];
        int32_t w = entry.data.i32[i + 1];
        int32_t h = entry.data.i32[i + 2];
        int32_t type = entry.data.i32[i + 3];
        Uint32 fmt;

        if (type == ACAMERA_SCALER_AVAILABLE_STREAM_CONFIGURATIONS_INPUT) {
            continue;
        }


        fmt = format_android_2_sdl(f);
        if (fmt != format) {
            continue;
        }

        if (i2 == index) {
            *width = w;
            *height = h;
            return 0;
        }

        i2++;
    }
    return -1;
}

int
GetDeviceName(int index, char *buf, int size)
{
    create_cameraMgr();

    if (cameraIdList == NULL) {
        GetNumDevices();
    }

    if (cameraIdList) {
        if (index >= 0 && index < cameraIdList->numCameras) {
            SDL_snprintf(buf, size, "%s", cameraIdList->cameraIds[index]);
            return 0;
        }
    }

    return -1;
}

int
GetNumDevices(void)
{
    camera_status_t res;
    create_cameraMgr();

    if (cameraIdList) {
        ACameraManager_deleteCameraIdList(cameraIdList);
        cameraIdList = NULL;
    }

    res = ACameraManager_getCameraIdList(cameraMgr, &cameraIdList);

    if (res == ACAMERA_OK) {
        if (cameraIdList) {
            return cameraIdList->numCameras;
        }
    }
    return -1;
}


#elif __IPHONEOS__ || __MACOS__
    /* SDL_video_capture_apple.m */
#else
int
OpenDevice(_THIS)
{
    return SDL_SetError("not implemented");
}

void
CloseDevice(_THIS)
{
    return;
}

int
InitDevice(_THIS)
{
    Uint32 sz, p;
    size_buffer(_this->spec.width, _this->spec.height, _this->spec.format, &sz, &p);
    SDL_Log("size_buffer: %" SDL_PRIu32 " x %" SDL_PRIu32 "", _this->spec.width, _this->spec.height);
    return -1;
}

int
GetDeviceSpec(_THIS, SDL_VideoCaptureSpec *spec)
{
    return -1;
}

int
StartCapture(_THIS)
{
    return -1;
}

int
StopCapture(_THIS)
{
    return -1;
}

int
AcquireFrame(_THIS, SDL_VideoCaptureFrame *frame, Uint64 *ticks)
{
    return -1;
}

int
ReleaseFrame(_THIS, SDL_VideoCaptureFrame *frame)
{
    return -1;
}

int
GetNumFormats(_THIS)
{
    return -1;
}

int
GetFormat(_THIS, int index, Uint32 *format, Uint32 *type)
{
    return -1;
}

int
GetNumFrameSizes(_THIS, Uint32 format)
{
    return -1;
}

int
GetFrameSize(_THIS, Uint32 format, int index, Uint32 *width, Uint32 *height)
{
    return -1;
}

int
GetDeviceName(int index, char *buf, int size)
{
    return -1;
}

int
GetNumDevices(void)
{
    return -1;
}
#endif
