/*
  Copyright (C) 1997-2022 Sam Lantinga <slouken@libsdl.org>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely.
*/
#include "SDL3/SDL.h"
#include "SDL3/SDL_video_capture.h"
#include <stdio.h>

/* Enable DMABUF to compile (linux + v4l2) */
#define USE_DMABUF 0

#if USE_DMABUF
#  include "SDL_egl.h"
#  include "SDL_opengles2.h"
#  include "drm/drm_fourcc.h"
#endif

#ifdef __EMSCRIPTEN__
#include <emscripten/emscripten.h>
#endif

static const char *usage = "\
 \n\
 =========================================================================\n\
 \n\
Use keyboards:\n\
 o: open first video capture device. (close previously opened)\n\
 l: list video capture devices\n\
 i: information about status (Init, Playing, Stopped)\n\
 f: formats and resolutions available\n\
 s: start / stop capture\n\
 h: display help\n\
 esc: exit \n\
 \n\
 =========================================================================\n\
 \n\
";

typedef struct {
    Uint64 next_check;
    int frame_counter;
    int check_delay;
    double last_fps;
} measure_fps_t;

static void
update_fps(measure_fps_t *m, const char *str)
{
    Uint64 now = SDL_GetTicks();
    Uint64 deadline;
    m->frame_counter++;
    if (m->check_delay == 0) {
        m->check_delay = 1500;
    }
    if (str == NULL) {
        str = "";
    }
    deadline = m->next_check;
    if (now >= deadline) {
        /* Print out some timing information */
        const Uint64 then = m->next_check - m->check_delay;
        m->last_fps = ((double) m->frame_counter * 1000) / (now - then);

        SDL_Log("%s %2.2f frames per second\n", str, m->last_fps);
        m->next_check = now + m->check_delay;
        m->frame_counter = 0;
    }
}

/*
 +--1--+
 |     |
 4     6
 |--2--|
 5     7
 |     |
 +--3--+
*/
static const
SDL_FPoint segments[9][2] = {
    { {0, 0},  {10, 1}  },  /* 1 */
    { {0, 10}, {10, 11} },  /* 2 */
    { {0, 20}, {10, 21} },  /* 3 */

    { {0, 0},  {1, 11}  },  /* 4 */
    { {0, 10}, {1, 21}  },  /* 5 */

    { {9, 0},  {10, 11} },  /* 6 */
    { {9, 10}, {10, 21} },  /* 7 */
    { {3, 18}, {5, 21}  },  /* . */
    { {4, 0},  {5, 21}  }   /* | */
};

static void
draw_string(SDL_Renderer *renderer, const char *c, float x0, float y0)
{
    while (*c) {
        int seg = 0;
        int val = *c;
        if (val >= 'A' && val <= 'Z') {
            val = 'a' + val - 'A';
        }
        switch (val) {
            case '0': seg = 167345; break;
            case '1': seg = 67; break;
            case '2': seg = 16253; break;
            case '3': seg = 16732; break;
            case '4': seg = 4267; break;
            case '5': seg = 14273; break;
            case '6': seg = 145372; break;
            case '7': seg = 167; break;
            case '8': seg = 1234567; break;
            case '9': seg = 123467; break;
            case '.': seg = 8; break;
            case '-': seg = 2; break;
            case '_': seg = 3; break;
            case '+': seg = 29; break;
            case '|': seg = 9; break;
            case '[': seg = 1453; break;
            case ']': seg = 1673; break;
            case 'a': seg = 124567; break;
            case 'b': seg = 1234567; break;
            case 'c': seg = 1453; break;
            case 'd': seg = 167345; break;
            case 'e': seg = 12345; break;
            case 'f': seg = 1425; break;
            case 'g': seg = 123457; break;
            case 'h': seg = 24567; break;
            case 'i': seg = 9; break;
            case 'j': seg = 6735; break;
            case 'k': seg = 24567; break;
            case 'l': seg = 453; break;
            case 'm': seg = 145679; break;
            case 'n': seg = 14567; break;
            case 'o': seg = 134567; break;
            case 'p': seg = 14256; break;
            case 'q': seg = 134567; break;
            case 'r': seg = 145; break;
            case 's': seg = 14273; break;
            case 't': seg = 19; break;
            case 'u': seg = 45367; break;
            case 'v': seg = 45367; break;
            case 'w': seg = 345679; break;
            case 'x': seg = 24567; break;
            case 'y': seg = 23467; break;
            case 'z': seg = 16253; break;
        }
        while (seg) {
            SDL_FRect r;
            SDL_FPoint p1, p2;
            int s = seg - 10 * (seg/10);
            p1 = segments[s - 1][0];
            p2 = segments[s - 1][1];
            r.x = p1.x + x0;
            r.y = p1.y + y0;
            r.w = p2.x - p1.x;
            r.h = p2.y - p1.y;
            SDL_RenderRect(renderer, &r);
            seg /= 10;
        }
        x0 += 12;
        c++;
    }
}

#if defined(__linux__) && !defined(__ANDROID__)
static void load_average(float *val)
{
    FILE *fp = 0;
    char line[1024];
    fp = fopen("/proc/loadavg", "rt");
    if (fp) {
        char *s = fgets(line, sizeof(line), fp);
        if (s) {
            SDL_sscanf(s, "%f", val);
        }
        fclose(fp);
    }
}
#endif

int main(int argc, char **argv)
{
    SDL_Window *window = NULL;
    SDL_Renderer *renderer = NULL;
    SDL_Texture *texture = NULL;
    int texture_updated = 0;
    SDL_Event evt;
    int quit = 1;

    SDL_VideoCaptureDeviceID id = 0;
    SDL_VideoCaptureSpec obtained;

    SDL_FRect r_playstop = { 50, 50, 120, 50 };
    SDL_FRect r_close = { 50 + (120 + 50) * 1, 50, 120, 50 };

    SDL_FRect r_open0 = { 50 + (120 + 50) * 2, 50, 120, 50 };
    SDL_FRect r_open1 = { 50 + (120 + 50) * 2, 150, 120, 50 };

    SDL_FRect r_format = { 50 + (120 + 50) * 3, 50, 120, 50 };
    SDL_FRect r_listdev = { 50 + (120 + 50) * 4, 50, 120, 50 };

    int start = 1;
    SDL_VideoCaptureFrame frame_current;

#if USE_DMABUF
    EGLDisplay display;

    typedef EGLDisplay (*eglGetCurrentDisplay_t)();
    typedef EGLImageKHR (*eglCreateImageKHR_t)(EGLDisplay dpy, EGLContext ctx, EGLenum target, EGLClientBuffer buffer, const EGLint *attrib_list);
    typedef EGLBoolean (*eglDestroyImageKHR_t)(EGLDisplay dpy, EGLImageKHR image);
    typedef void (*glActiveTexture_t)(GLenum texture);
    typedef void (*glEGLImageTargetTexture2DOES_t)(GLenum target, GLeglImageOES image);
    typedef EGLint (*eglGetError_t)(void);

    eglGetCurrentDisplay_t eglGetCurrentDisplay;
    eglCreateImageKHR_t eglCreateImageKHR;
    eglDestroyImageKHR_t eglDestroyImageKHR;
    glActiveTexture_t glActiveTexture;
    glEGLImageTargetTexture2DOES_t glEGLImageTargetTexture2DOES;
    eglGetError_t eglGetError;

#endif

    measure_fps_t fps_main;
    measure_fps_t fps_capture;

    SDL_zero(fps_main);
    SDL_zero(fps_capture);
    SDL_zero(frame_current);

    SDL_Log("%s", usage);

    /* Enable standard application logging */
    SDL_LogSetPriority(SDL_LOG_CATEGORY_APPLICATION, SDL_LOG_PRIORITY_INFO);

#if USE_DMABUF
    SDL_SetHint("SDL_RENDER_OPENGL_NV12_RG_SHADER", "1");
    SDL_SetHint("SDL_VIDEO_X11_FORCE_EGL", "1"); /* Don't use GLX */
    SDL_SetHint(SDL_HINT_RENDER_DRIVER, "opengles2");
#endif

    /* Load the SDL library */
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Couldn't initialize SDL: %s", SDL_GetError());
        return 1;
    }

    if (!(window = SDL_CreateWindow("Local Video", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 1000, 800, 0))) {
        SDL_Log("Couldn't create window: %s", SDL_GetError());
        return 1;
    }

    SDL_LogSetAllPriority(SDL_LOG_PRIORITY_VERBOSE);

    renderer = SDL_CreateRenderer(window, NULL, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
    if (renderer == NULL) {
        SDL_Log("Couldn't create renderer: %s", SDL_GetError());
        return 1;
    }

#if USE_DMABUF

#define LOAL_FUNC(f)                            \
    f = (f ##_t)SDL_GL_GetProcAddress(#f);      \
    if (f == NULL) {                            \
        SDL_Log("Cannot load function: " #f );  \
        return 0;                               \
    }                                           \

    LOAL_FUNC(eglGetError)
    LOAL_FUNC(eglGetCurrentDisplay)
    LOAL_FUNC(eglCreateImageKHR)
    LOAL_FUNC(eglDestroyImageKHR)
    LOAL_FUNC(glActiveTexture)
    LOAL_FUNC(glEGLImageTargetTexture2DOES)

    display = eglGetCurrentDisplay();
    if (display == EGL_NO_DISPLAY) {
        SDL_Log("no dispay");
        return 0;
    }
#endif


    SDL_LogSetAllPriority(SDL_LOG_PRIORITY_VERBOSE);

    id = SDL_OpenVideoCapture(NULL);

    if (id == SDL_VIDEO_CAPTURE_INVALID) {
        SDL_Log("Error SDL_OpenVideoCapture: %s", SDL_GetError());
    }

    {
        /* List formats */
        int i, num = SDL_GetNumVideoCaptureFormats(id);
        for (i = 0; i < num; i++) {
            Uint32 format, type;
            SDL_GetVideoCaptureFormat(id, i, &format, &type);
            SDL_Log("format %d/%d type=%08" SDL_PRIx32 ": %s", i, num, type, SDL_GetPixelFormatName(format));
            {
                Uint32 w, h;
                int j, num2 = SDL_GetNumVideoCaptureFrameSizes(id, format);
                for (j = 0; j < num2; j++) {
                    SDL_GetVideoCaptureFrameSize(id, format, j, &w, &h);
                    SDL_Log("  framesizes %d/%d :  %" SDL_PRIu32 " x %" SDL_PRIu32 "", j, num2, w, h);
                }
            }
        }
    }

    /* Set Spec */
    {
        int ret;
        /* forced_format */
        SDL_VideoCaptureSpec desired;
        SDL_zero(desired);
        desired.width = 640 * 2;
        desired.height = 360 * 2;
        ret = SDL_SetVideoCaptureSpec(id, &desired, &obtained, SDL_VIDEO_CAPTURE_ALLOW_ANY_CHANGE);

        if (ret < 0) {
            SDL_SetVideoCaptureSpec(id, NULL, &obtained, 0);
        }
    }

    SDL_Log("Open capture video device. Obtained spec: size=%" SDL_PRIu32 "x%" SDL_PRIu32 " format=%s",
            obtained.width, obtained.height, SDL_GetPixelFormatName(obtained.format));

    {
        SDL_VideoCaptureSpec spec;
        if (SDL_GetVideoCaptureSpec(id, &spec) == 0) {
            SDL_Log("Read spec: size=%" SDL_PRIu32 "x%" SDL_PRIu32 " format=%s",
                    spec.width, spec.height, SDL_GetPixelFormatName(spec.format));
        } else {
            SDL_Log("Error read spec: %s", SDL_GetError());
        }
    }

    if (SDL_StartVideoCapture(id) < 0) {
        SDL_Log("error SDL_StartVideoCapture(): %s", SDL_GetError());
    }

    while (quit) {

        SDL_SetRenderDrawColor(renderer, 0x99, 0x99, 0x99, 255);
        SDL_RenderClear(renderer);

        SDL_SetRenderDrawColor(renderer, 0x33, 0x33, 0x33, 255);

        SDL_RenderFillRect(renderer, &r_playstop);
        SDL_RenderFillRect(renderer, &r_close);
        SDL_RenderFillRect(renderer, &r_open0);
        SDL_RenderFillRect(renderer, &r_open1);
        SDL_RenderFillRect(renderer, &r_format);
        SDL_RenderFillRect(renderer, &r_listdev);

        SDL_SetRenderDrawColor(renderer, 0xcc, 0xcc, 0xcc, 255);

        draw_string(renderer, "play stop", r_playstop.x + 5, r_playstop.y + 5);
        draw_string(renderer, "close", r_close.x + 5, r_close.y + 5);
        draw_string(renderer, "open dev0", r_open0.x + 5, r_open0.y + 5);
        draw_string(renderer, "open dev1", r_open1.x + 5, r_open1.y + 5);
        draw_string(renderer, "formats", r_format.x + 5, r_format.y + 5);
        draw_string(renderer, "devices", r_listdev.x + 5, r_listdev.y + 5);



        while (SDL_PollEvent(&evt) == 1) {
            int sym = 0;
            int opendev = 0;
            switch (evt.type)
            {
                case SDL_EVENT_KEY_DOWN:
                    {
                        sym = evt.key.keysym.sym;
                        break;
                    }
                case SDL_EVENT_QUIT:
                    {
                        quit = 0;
                        SDL_Log("Ctlr+C : Quit!");
                    }
                    break;

                case SDL_EVENT_MOUSE_BUTTON_DOWN:
                    {
                        SDL_FRect *r = NULL;
                        SDL_FPoint pt;
                        pt.x = evt.button.x;
                        pt.y = evt.button.y;
                        if (SDL_PointInRectFloat(&pt, &r_playstop)) {
                            r = &r_playstop;
                            sym = SDLK_s;
                        }
                        if (SDL_PointInRectFloat(&pt, &r_close)) {
                            r = &r_close;
                            sym = SDLK_c;
                        }
                        if (SDL_PointInRectFloat(&pt, &r_open0)) {
                            r = &r_open0;
                            sym = SDLK_o;
                            opendev = 0;
                        }

                        if (SDL_PointInRectFloat(&pt, &r_open1)) {
                            r = &r_open1;
                            sym = SDLK_o;
                            opendev = 1;
                        }

                        if (SDL_PointInRectFloat(&pt, &r_format)) {
                            r = &r_format;
                            sym = SDLK_f;
                        }
                        if (SDL_PointInRectFloat(&pt, &r_listdev)) {
                            r = &r_listdev;
                            sym = SDLK_l;
                        }


                        if (r) {
                            SDL_SetRenderDrawColor(renderer, 0x33, 0, 0, 255);
                            SDL_RenderFillRect(renderer, r);
                        }
                    }

                    break;
            }

            if (sym == SDLK_c) {
                if (frame_current.num_planes) {
                    SDL_VideoCaptureReleaseFrame(id, &frame_current);
                }
                SDL_CloseVideoCapture(id);
                id = SDL_VIDEO_CAPTURE_INVALID;
                SDL_Log("Close");
            }

            if (sym == SDLK_o) {
                if (id != SDL_VIDEO_CAPTURE_INVALID) {
                    SDL_Log("Close previous ..");
                    if (frame_current.num_planes) {
                        SDL_VideoCaptureReleaseFrame(id, &frame_current);
                    }
                    SDL_CloseVideoCapture(id);
                }

                texture_updated = 0;

                SDL_ClearError();

                if (opendev == 0) {
                    id = SDL_OpenVideoCaptureWithSpec(NULL, &obtained, &obtained, SDL_VIDEO_CAPTURE_ALLOW_ANY_CHANGE);
                } else {
                    SDL_Log("Try to open:%s", SDL_GetVideoCaptureDeviceName(opendev));
                    id = SDL_OpenVideoCaptureWithSpec(SDL_GetVideoCaptureDeviceName(opendev), &obtained, &obtained, SDL_VIDEO_CAPTURE_ALLOW_ANY_CHANGE);
                }

                /* spec may have changed because of re-open */
                if (texture) {
                    SDL_DestroyTexture(texture);
                    texture = NULL;
                }

                SDL_Log("Open id:%" SDL_PRIu32 " %s", id, SDL_GetError());
                start = 0;
            }

            if (sym == SDLK_l) {
                int num = SDL_GetNumVideoCaptureDevices();
                int i;
                SDL_Log("Num devices : %d", num);
                for (i = 0; i < num; i++) {
                    SDL_Log("Device %d/%d : %s", i, num, SDL_GetVideoCaptureDeviceName(i));
                }
            }

            if (sym == SDLK_i) {
                SDL_VideoCaptureStatus status = SDL_GetVideoCaptureStatus(id);
                if (status == SDL_VIDEO_CAPTURE_STOPPED) SDL_Log("STOPPED");
                if (status == SDL_VIDEO_CAPTURE_PLAYING) SDL_Log("PLAYING");
                if (status == SDL_VIDEO_CAPTURE_INIT) SDL_Log("INIT");
            }

            if (sym == SDLK_s) {
                start = !start;

                if (start) {
                    SDL_Log("Start");
                    SDL_StartVideoCapture(id);
                } else {
                    SDL_Log("Stop");
                    SDL_StopVideoCapture(id);
                }

            }

            if (sym == SDLK_f) {
                SDL_Log("List formats");

                if (id == SDL_VIDEO_CAPTURE_INVALID) {
                    id = SDL_OpenVideoCapture(NULL);
                }

                /* List formats */
                {
                    int i, num = SDL_GetNumVideoCaptureFormats(id);
                    for (i = 0; i < num; i++) {
                        Uint32 format, type;
                        SDL_GetVideoCaptureFormat(id, i, &format, &type);
                        SDL_Log("format %d/%d type=%08" SDL_PRIx32 ": %s", i, num, type, SDL_GetPixelFormatName(format));
                        {
                            Uint32 w, h;
                            int j, num2 = SDL_GetNumVideoCaptureFrameSizes(id, format);
                            for (j = 0; j < num2; j++) {
                                SDL_GetVideoCaptureFrameSize(id, format, j, &w, &h);
                                SDL_Log("  framesizes %d/%d :  %" SDL_PRIu32 " x %" SDL_PRIu32 "", j, num2, w, h);
                            }
                        }
                    }
                }
            }
            if (sym == SDLK_ESCAPE || sym == SDLK_AC_BACK) {
                quit = 0;
                SDL_Log("Key : Escape!");
            }

            if (sym == SDLK_h || sym == SDLK_F1) {
                SDL_Log("%s", usage);
            }
        }

        if (id == SDL_VIDEO_CAPTURE_INVALID) {
            /* device has been closed */
            frame_current.num_planes = 0;
            texture_updated = 0;
        } else {
            Uint64 ticks;
            int ret;
            SDL_VideoCaptureFrame frame_next;
            SDL_zero(frame_next);

            ret = SDL_VideoCaptureAcquireFrame(id, &frame_next, &ticks);
            if (ret < 0) {
                SDL_Log("err SDL_VideoCaptureAcquireFrame: %s", SDL_GetError());
            }
#if 1
            if (frame_next.num_planes) {
                SDL_Log("frame: %p  at %" SDL_PRIu64, (void*)frame_next.data[0], ticks);
            }
#endif

            if (frame_next.num_planes) {

                update_fps(&fps_capture, "Capture");

                if (frame_current.num_planes) {
                    ret = SDL_VideoCaptureReleaseFrame(id, &frame_current);
                    if (ret < 0) {
                        SDL_Log("err SDL_VideoCaptureReleaseFrame: %s", SDL_GetError());
                    }
                }
                frame_current = frame_next;
                texture_updated = 0;
            }
        }

        /* Moving square */
        SDL_SetRenderDrawColor(renderer, 0, 0xff, 0, 255);
        {
            SDL_FRect r;
            static float x = 0;
            x += 10;
            if (x > 1000) {
                x = 0;
            }
            r.x = x;
            r.y = 100;
            r.w = r.h = 10;
            SDL_RenderFillRect(renderer, &r);
        }

        SDL_SetRenderDrawColor(renderer, 0x33, 0x33, 0x33, 255);

        /* Update SDL_Texture with last video frame (only once per new frame) */
        if (frame_current.num_planes && texture_updated == 0) {
#if USE_DMABUF
            int use_sw = 0;
#else
            int use_sw = 1;
#endif

            /* Create texture with appropriate format (for DMABUF or not) */
            if (texture == NULL) {
                Uint32 format = obtained.format;
#if USE_DMABUF
                if ((frame_current.fd == -1 && frame_current.clientbuffer == NULL) || use_sw) {
                    /* keep format */
                } else {
                    format = SDL_PIXELFORMAT_EXTERNAL_OES;
                }
#endif
                texture = SDL_CreateTexture(renderer, format, SDL_TEXTUREACCESS_STATIC, obtained.width, obtained.height);
                if (texture == NULL) {
                    SDL_Log("Couldn't create texture: %s", SDL_GetError());
                    return 1;
                }
            }

            if ((frame_current.fd == -1 && frame_current.clientbuffer == NULL) || use_sw) {
                /* Use software data */
                if (frame_current.num_planes == 1) {
                    SDL_UpdateTexture(texture, NULL,
                            frame_current.data[0], frame_current.row_stride[0]);
                } else if (frame_current.num_planes == 2) {
                    SDL_UpdateNVTexture(texture, NULL,
                            frame_current.data[0], frame_current.row_stride[0],
                            frame_current.data[1], frame_current.row_stride[1]);
                } else if (frame_current.num_planes == 3) {
                    SDL_UpdateYUVTexture(texture, NULL, frame_current.data[0], frame_current.row_stride[0],
                            frame_current.data[1], frame_current.row_stride[1],
                            frame_current.data[2], frame_current.row_stride[2]);
                }
                texture_updated = 1;
            } else {
                /* Use DMABUF */
#if USE_DMABUF
                int j = 0;
                EGLImage image;
                EGLint img_attr[64];

#if __ANDROID__
                img_attr[j] = EGL_NONE;
                image = eglCreateImageKHR(display, EGL_NO_CONTEXT, EGL_NATIVE_BUFFER_ANDROID, frame_current.clientbuffer, img_attr);
#else
                /* V4L2_PIX_FMT_YUYV / SDL_PIXELFORMAT_EXTERNAL_OES */
                img_attr[j++] = EGL_WIDTH;
                img_attr[j++] = obtained.width;

                img_attr[j++] = EGL_HEIGHT;
                img_attr[j++] = obtained.height;

                img_attr[j++] = EGL_LINUX_DRM_FOURCC_EXT;
                img_attr[j++] = DRM_FORMAT_YUYV;

                img_attr[j++] = EGL_DMA_BUF_PLANE0_FD_EXT;
                img_attr[j++] = frame_current.fd;
                img_attr[j++] = EGL_DMA_BUF_PLANE0_OFFSET_EXT;
                img_attr[j++] = 0;
                img_attr[j++] = EGL_DMA_BUF_PLANE0_PITCH_EXT;
                img_attr[j++] = frame_current.row_stride[0];

                img_attr[j++] = EGL_NONE;

                image = eglCreateImageKHR(display, EGL_NO_CONTEXT, EGL_LINUX_DMA_BUF_EXT, NULL, img_attr);
#endif

                if (image == EGL_NO_IMAGE_KHR) {
                    SDL_Log("error eglCreateImageKHR : eglGetError: %08" SDL_PRIx32 "", eglGetError());
                } else {
                    SDL_Log("ImageKHR: %p fd: %" SDL_PRIu32 "", image, frame_current.fd);

                    SDL_GL_BindTexture(texture, NULL, NULL);
                    glActiveTexture(GL_TEXTURE0);
                    glEGLImageTargetTexture2DOES(GL_TEXTURE_EXTERNAL_OES, image);
                    SDL_GL_UnbindTexture(texture);

                    eglDestroyImageKHR(display, image);
                    texture_updated = 1;
                }
#endif
            }
        }

        /* RenderCopy the SDL_Texture */
        if (texture_updated == 1) {
#if 0
            SDL_Rect dst;
            dst.x = 10;
            dst.y = 120;
            dst.w = obtained.width;
            dst.h = obtained.height;

            SDL_RenderCopy(renderer, texture, NULL, &dst);
#else
            { /* Scale texture to fit the screen */
                int w, h;
                int tw, th;
                SDL_FRect d;
                SDL_GetWindowSize(window, &w, &h);
                SDL_QueryTexture(texture, NULL, NULL, &tw, &th);

                if (tw > w - 20) {
                    float scale = (float) (w - 20) / (float) tw;
                    tw = w - 20;
                    th = (int)((float) th * scale);
                }
                d.x = 10;
                d.y = (float)(h - th);
                d.w = (float)tw;
                d.h = (float)(th - 10);
                SDL_RenderTexture(renderer, texture, NULL, &d);
            }
#endif
        }

        /* display FPS */
        {
            char buf[128];
            SDL_snprintf(buf, sizeof(buf), "%2.2f fps", fps_capture.last_fps);
            draw_string(renderer, buf, 10, 10);
        }

        /* display status */
        {
            const char *status = "no status";
            if (id != SDL_VIDEO_CAPTURE_INVALID) {
                SDL_VideoCaptureStatus s = SDL_GetVideoCaptureStatus(id);
                if (s == SDL_VIDEO_CAPTURE_INIT) {
                    status = "init";
                } else if (s == SDL_VIDEO_CAPTURE_PLAYING) {
                    status = "playing";
                } else if (s == SDL_VIDEO_CAPTURE_STOPPED) {
                    status = "stopped";
                }
            }
            draw_string(renderer, status, 200, 10);
        }

        /* display last error */
        {
            draw_string(renderer, SDL_GetError(), 400, 10);
        }

        /* display spec */
        {
            char buf[128];
            SDL_snprintf(buf, sizeof(buf), "%" SDL_PRIu32 " x %" SDL_PRIu32 "  %s", obtained.width, obtained.height, SDL_GetPixelFormatName(obtained.format));
            draw_string(renderer, buf, 10, 200);
        }

        /* display load average */
#if defined(__linux__) && !defined(__ANDROID__)
        {
            float val = 0.0f;
            char buf[128];
            load_average(&val);
            if (val != 0.0f) {
                SDL_snprintf(buf, sizeof(buf), "load avg %2.2f percent", val);
                draw_string(renderer, buf, 800, 10);
            }
        }
#endif


        SDL_RenderPresent(renderer);

        update_fps(&fps_main, "Main");

    }

    if (SDL_StopVideoCapture(id) < 0) {
        SDL_Log("error SDL_StopVideoCapture(): %s", SDL_GetError());
    }

    if (frame_current.num_planes) {
        SDL_VideoCaptureReleaseFrame(id, &frame_current);
    }

    SDL_CloseVideoCapture(id);

    if (texture) SDL_DestroyTexture(texture);
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);

    SDL_Quit();

    return 0;
}
